#!/usr/local/bin/ruby
# $Id: cutil.rb,v 1.24 2002/10/30 07:38:39 abe Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   Corpus Utility
#
#      Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>
######################################################################
require File::join( File.dirname(__FILE__), "nlputil.rb" )


#
#   Global Method
######################################################################

# extract [hp̃ftHg萔
MODEextract_DefaultCount = 100

$VERSION = <<"--------"
#{GEMINI::mklogo("Corpus Utility   ver. 2.0")}
        Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>

--------

$USAGE_MAIN = <<"--------"
usage:
  cutil.rb [mode] (option)

mode:
  extract           Extract sentences from corpora
  divide            Divide corpora
  select            Extract data blocks from a corpus file
  count             Count sentences in corpora
  sample            Extract specified sentences from a corpus
  unite             Unite some corpora
  analyze           Analyze corpora and output result
  merge             Merge source corpus into destination corpus
             (To read more information, type as 'cutil.rb -h extract')

--------

$USAGE_MODE =
{
'extract' => <<"--------",
extract mode usage:
  cutil.rb extract (options) [corpus files]

extract mode options:
  --count COUNT               extract count (default=#{MODEextract_DefaultCount})
  -c COUNT                      (be padded if data is not enough)

  --nocopy                    stop coping (padding) sentences
--------

'divide' => <<"--------",
divide mode usage:
  cutil.rb divide -p PREFIX (options) [corpus files]

divide mode options:
  --prefix PREFIX             prefix name for divided corpus *REQUIRED*
  -p PREFIX

  --width WIDTH               collect sentences according to width
  -w WIDTH                      (collides with -c and --count)
                              e.g. -w 2  =  1,2   ; 3,4   ; ...
                                   -w 3  =  1,2,3 ; 4,5,6 ; ...

  --count COUNT               divide given corpora into N corpora
  -c COUNT                      (collides with -w and --width)
--------

'select' => <<"--------",
select mode usage:
  cutil.rb select [corpus file] [block names]
--------

'count' => <<"--------",
count mode usage:
  cutil.rb count [corpus files]
--------

'sample' => <<"--------",
sample mode usage:
  cutil.rb sample [corpus file] [spec]

sample mode examples:
  e.g. 1,2,3 = [1], [2] and [3]
       1-4,7 = [1], [2], [3], [4] and [7]
--------

'unite' => <<"--------",
unite mode usage:
  cutil.rb unite [corpus files]
--------

'analyze' => <<"--------",
analyze mode usage:
  cutil.rb analyze [corpus files]

analyze mode options:
  --simply                    output result only.
--------

'merge' => <<"--------",
merge mode usage:
  cutil.rb merge [source corpus] [destination corpus]
--------
}

$USAGE = lambda do
  # ̃ubN]Ƃ $files Ƀt@C񂪊i[Ă
  $USAGE_MAIN + String($USAGE_MODE[$files.first])
end



#
#   extract
######################################################################

def extract_process ( corpus, buffer )
  begin
    v_print "   - ID = #{$_}\n" if $_ = corpus['id']
    readcount = 0
    t = 0
    corpus.each do |sentence|
      t = update_progress_info( t ) { v_print "     - #{readcount}\r" }
      $allcount += 1
      readcount += 1
      if buffer.size >= $Count and rand($allcount) != 0 then
        sentence.close()
        next
      end
      
      image = ""
      sentence.dump( image )
      
      if buffer.size < $Count then
        buffer << image
      else
        buffer[ rand( buffer.size ) ] = image
      end
    end
    v_print "     - #{readcount}\n"
  rescue
    fatal_exit( root, $! )
  end
end

def mode_extract ()
  v_print " + MODE = extract\n"
  
  $Count = Integer( $OPT_count || MODEextract_DefaultCount )
  v_print " + extract-size = #{$Count}\n"
  
  # buffer ɕf[^~ςĂ
  buffer = [ ]
  corpus_prop = nil
  $allcount = 0
  
  $files.each do |filename|
    stream = filename_to_stream( filename )
    next unless stream
    
    CorpusStream.new( stream ) do |root|
      root.each do |corpus|
        corpus_prop = corpus.block_property unless corpus_prop
        extract_process( corpus, buffer )
      end
    end
  end
  
  # Ȃꍇ͖ߍ킹
  if !defined?($OPT_nocopy) or !$OPT_nocopy then
    while buffer.size < $Count do
      $_ = rand( buffer.size )
      buffer << buffer[ $_ ]
    end
  end
  display_block(CorpusBlock, corpus_prop) do
    buffer.each {|item| $defout << item }
  end
  v_print " + OK.\n"
end


#
#   divide
######################################################################

# Ruby 1.6.5 Ȃǂł Range#eql? ɃoÔ Range ͎gȂ
def  length_to_range  ( len )
  # e.g.  ($Width == 4)
  #  1 --> [1,4], 2 --> [1,4]
  #  7 --> [5,8], 10 --> [9,12]
  if len == 0 then
    return [0, 0]
  end
  block = Integer((len-1) / $Width)
  return [ block*$Width + 1, (block+1)*$Width ]
end

def mode_divide_width ()
  $Width = Integer($OPT_width)
  fatal_exit "Invalid -w argument." if $Width <= 0
  v_print " + width = #{$Width}\n"
  
  buffer    = [ ]
  container = { }  #  range ɑΉ IO IuWFNg ( IO IuWFNg̔z)
  readcount = 0
  
  corpus_prop = nil
  
  $files.each do |filename|
    stream = filename_to_stream( filename )
    next unless stream
    
    CorpusStream.new( stream ) do |root|
      begin
        root.each do |corpus|
          corpus_prop = Block::mkstr_property(corpus.block_property) unless corpus_prop
          v_print "   - ID = #{$_}\n" if $_
          
          readcount = 0
          t = 0
          corpus.each do |sentence|
            t = update_progress_info( t ) { v_print "     - #{readcount}\r" }
            readcount += 1
            
            # ܂͕ɓǂݍ (ǂݏIȂƕ킩炸o͐킩Ȃ)
            image    = [ ]
            keyrange = nil
            sentence.out = image
            sentence.display( image ) do
              sentence.each( PoSBlock ) do |cls, block|
                if cls then
                  block.display { keyrange = length_to_range( cls.new(block).read.length ) }
                else
                  block.dump( image )
                end
              end
            end
            unless keyrange then
              print_warn( stream, "\##{readcount}:No <dPoS>. (lost data)" )
              next
            end
            
            # t@Cւ̏o
            file = container[ keyrange ]
            unless file then
              if keyrange.first != keyrange.last then
                suffix = "#{keyrange.first}-#{keyrange.last}"
              else
                suffix = "#{keyrange.first}"
              end
              out_filename = "#{$Prefix}_#{suffix}.corpus"
              file = container[ keyrange ] = open( out_filename, "w" )
              
              if corpus_prop.empty? then
                file << "<pCorpus>\n"
              else
                file << "<pCorpus #{corpus_prop}>\n"
              end
            end
            for line in image do # image ͍sƂ̔z
              file << line
            end
          end
          v_print "     - #{readcount}\n"
        end
      rescue
        fatal_exit( root, $! )
      end
    end
  end
  
  v_print " + written files:\n"
  container.each_pair do |range, file|
    v_print "   - #{file.path}\n"
    file << "</pCorpus>\n"
    file.close
  end
end

def mode_divide_count ()
  $Count = Integer($OPT_count)
  fatal_exit "Invalid -c argument." if $Count <= 1
  v_print " + count = #{$Count}\n"
  
  container = Array.new( $Count ) # IO IuWFNgi[
  numberary = Array.new( $Count ) # U蕪p̃CfNXXbgĂ
  for i in 0...$Count do
    container[i] = "#{$Prefix}_#{i+1}.corpus"
    numberary[i] = i
  end
  schedule = [ ]
  
  corpus_prop = nil
  
  readcount = 0
  $files.each do |filename|
    stream = filename_to_stream( filename )
    next unless stream
    
    CorpusStream.new( stream ) do |root|
      begin
        root.each do |corpus|
          $_ = corpus['id']
          corpus_prop = Block::mkstr_property(corpus.block_property) unless corpus_prop
          v_print "   - ID = #{$_}\n" if $_
          
          readcount = 0
          t = 0
          corpus.each do |sentence|
            t = update_progress_info( t ) { v_print "     - #{readcount}\r" }
            readcount += 1
            
            if schedule.empty? then # U蕪悪Ȃ̂Ŏ񕪂쐬
              schedule.replace( numberary )
              ($Count * 50).times do #  (50 ͏oC)
                a = rand( schedule.size )
                b = rand( schedule.size )
                $_ = schedule[a]; schedule[a] = schedule[b]; schedule[b] = $_
              end
            end
            
            out = container[ schedule.last ]
            if out.is_a?(String) then
              out = open( out, "w" )
              if corpus_prop.empty? then
                file << "<pCorpus>\n"
              else
                file << "<pCorpus #{corpus_prop}>\n"
              end
              container[ schedule.last ] = out
            end
            sentence.dump(out)
            schedule.pop
          end
          v_print "     - #{readcount}\n"
        end
      rescue
        fatal_exit( root, $! )
      end
    end
  end
  
  v_print " + written files:\n"
  for i in 0...$Count do
    file = container[i]
    v_print "   - #{file.path}\n"
    file << "</pCorpus>\n"
    file.close
  end
end

def mode_divide ()
  v_print " + MODE = divide\n"
  
  fatal_exit( "No prefix. Use option -p." ) unless $OPT_prefix
  
  $Prefix = $OPT_prefix
  $Width  = 1
  $Count  = nil
  v_print " + Prefix = #{$Prefix}\n"
  
  # -w  -c ͔rI
  fatal_exit( "-w and -c are exclusive." ) if $OPT_width and $OPT_count
  
  if $OPT_width then
    mode_divide_width()
  elsif $OPT_count then
    mode_divide_count()
  else
    fatal_exit("No instruction. Use -w or -c.")
  end
end


#
#   Select
######################################################################

def mode_select ()
  v_print " + MODE = select\n"
  targetblock = $files
  
  stream = filename_to_stream( $files.shift )
  fatal_exit("Select corpus.") unless stream
  
  CorpusStream.new( stream ) do |root|
    fatal_exit("No target block.") if targetblock.empty?
    v_print " + target-block = #{targetblock.join(', ')}\n" \
            "   - #{stream.path}\n"
    begin
      readcount = 0
      root.each do |corpus|
        $_ = corpus['id']
        v_print "   - ID = #{$_}\n" if $_
        
        t = 0
        corpus.display do
          corpus.each do |sentence|
            t = update_progress_info( t ) { v_print "     - #{readcount}\r" }
            readcount += 1
            sentence.display do
              sentence.each do |cls, block|
                if targetblock.include?( block.name ) then
                  block.dump( $defout )
                else
                  block.close()
                end
              end
            end
          end
        end
        v_print "     - #{readcount}\n"
      end
    rescue
      fatal_exit( root, $! )
    end
  end
end


#
#   Count
######################################################################

def mode_count ()
  v_print " + MODE = count\n"
  
  result = Hash.new( 0 )
  $files.each do |filename|
    stream = filename_to_stream( filename )
    next unless stream
    root = CorpusStream.new( stream )
    begin
      root.each do |corpus|
        readcount = 0
        corpus.each do |sentence|
          readcount += 1
        end
        v_print "   - #{readcount}\n"
        result[ filename ] += readcount
      end
    rescue
      fatal_exit( root, $! )
    ensure
      root.close
    end
  end
  result.each_pair do |filename, n|
    $defout << "#{filename.path}\t#{n}\n"
  end
end


#
#   Sample
######################################################################

class Integer
  def include? ( n )
    self == n
  end
end

def sample_spec_check ( n )
  $spec.each do |range|
    return true if range.include?( n )
  end
  return false
end

def sample_read_spec ( ary )
  ret = [ ]
  re_comma = /\s*,\s*/
  
  list = [ ]
  ary.each do |s|
    list.concat( s.split( re_comma ) )
  end
  list.each do |s|
    case s
    when /^(\d+)-(\d+)$/ then
      ret << ($1.to_i .. $2.to_i)
    when /^\d+$/ then
      ret << $&.to_i
    else
      fatal_exit("spec ``#{s}'' is invalid format.") unless s.empty?
    end
  end
  return ret
end

def sample_max_spec ()
  maxnum = 0
  $spec.each do |range|
    if range.is_a?( Range ) then
      maxnum = range.last if maxnum < range.last
    else
      maxnum = range if maxnum < range
    end
  end
  return maxnum
end

def mode_sample ()
  v_print " + MODE = sample\n"
  
  stream = filename_to_stream( $files.shift )
  fatal_exit("Select corpus.") unless stream
  
  $spec  = sample_read_spec( $files )
  maxnum = sample_max_spec()
  
  v_print " + spec = #{$spec.join(', ')}\n" \
          " + maxnum = #{maxnum}\n" \
          "   - #{stream.path}\n"
  
  CorpusStream.new( stream ) do |root|
    begin
      readcount = 0
      root.each_corpus do |corpus|
        t = 0
        corpus.display do
          corpus.each do |sentence|
            t = update_progress_info( t ) { v_print "     - #{readcount}\r" }
            readcount += 1
            if sample_spec_check( readcount ) then
              sentence.dump
            end
            break if readcount >= maxnum
          end
        end
      end
      v_print "     - #{readcount}\n"
    rescue
      fatal_exit( root, $! )
    end
  end
end


#
#   Unite
######################################################################

def unite_process ( corpus )
  begin
    t = 0
    readcount = 0
    corpus.each do |sentence|
      readcount += 1
      sentence.dump
      t = update_progress_info( t ) { v_print "      - #{readcount}\r" }
    end
    v_print "      - #{readcount}\n"
  rescue
    fatal_exit( root, $! )
  end
  return readcount
end

def mode_unite ()
  v_print " + MODE = unite\n"
  
  corpusprop = nil
  prefetch = find_first_corpus( $files )
  if prefetch then
    $files.unshift( prefetch )
    corpusprop = prefetch.corpus.block_property
  end
  
  totalcount = 0
  display_block(CorpusBlock, corpusprop) do
    $files.each do |filename|
      if String === filename then
        stream = filename_to_stream( filename )
        next unless stream
        root = CorpusStream.new(stream)
      else
        root, corpus = Array(filename)
        totalcount += unite_process( corpus )
      end
      
      root.each {|corpus| totalcount += unite_process(corpus) }
      root.close
    end
  end
  v_print " + total = #{totalcount}\n"
end


#
#   Analyze
######################################################################

def disp_block ( block )
  ($OPT_simply)? yield: block.display{ yield }
end

def mode_analyze ()
  v_print " + MODE = analyze\n"
  $grammar = Ext::Grammar.new()
  
  totalcount = 0
  totaltree = 0 ; count_tree = 0
  totallen  = 0 ; count_len  = 0
  $files.each do |filename|
    stream = filename_to_stream( filename )
    next unless stream
    CorpusStream.new(stream) do |root|
      begin
        root.each do |corpus|
          corpus.display( ($OPT_simply)? nil: $defout ) do
            t = 0
            readcount = 0
            corpus.each do |sentence|
              readcount += 1
              
              treecount = logblock = sentlen = nil
              sentence.out = $defout
              sentence.display do
                sentence.each( ItemSetBlock ) do |cls, block|
                  if cls then
                    $_ = disp_block(block) { cls.new(block).read() }
                    treecount, sentlen = $_.count_tree(), $_.length
                  else
                    if LogBlock === block then
                      block.out = nil
                      logblock = LogBlock.new() unless logblock
                      logblock.update( str_to_hash( block.read ) )
                    else
                      ($OPT_simply)? block.skip: block.dump
                    end
                  end
                end
              end
              
              if treecount then
                count_tree += 1
                totaltree += treecount
              end
              if sentlen then
                count_len += 1
                totallen += sentlen
              end
              unless $OPT_simply then
                logblock = LogBlock.new() unless logblock
                logblock['tree-count'] = treecount if treecount
                logblock['sentence-length'] = sentlen
                logblock.display()
              end
              
              t = update_progress_info( t ) { v_print "      - #{readcount}\r" }
            end
            v_print "      - #{readcount}\n"
            totalcount += readcount
          end
        end
      rescue
        fatal_exit( root, $! )
      end
    end
  end
  v_print " + total = #{totalcount}\n"
  
  logblock = LogBlock.new()
  logblock['tree-count'] = "#{totaltree/count_tree} # trees/sentence" if count_tree > 0
  logblock['sentence-length'] = sprintf( "%.2f # words/sentence", Float(totallen) / count_len ) if count_len > 0
  logblock.display()
end


#
#   Merge
######################################################################

class MergeMode
  def self.read_sentence ( corpus )
    (corpus.closed?)? nil: corpus.read()
  end
  
  def self.search_cache ( sent, cache )
    $_ = sent['id']
    return ($_)? cache[$_]: nil
  end
  
  Cache = Struct.new( :sentblock, :table, :id )
  def self.cache_sentence ( sent )
    table = { }
    sent.each_element() do |cls, block|
      image = ""
      block.dump( image )
      table[block.name] = image
    end
    return Cache.new(sent, table, 0)
  end
  
  def self.merge_src_block ( srcsent, table )
    srcsent.each_element() do |cls, block|
      table[block.name] = true
      block.dump
    end
  end
  
  def self.merge_dest_block ( destsent, table )
    destsent.each_element() do |cls, block|
      if table.include?(block.name) then
        block.skip
      else
        block.dump
      end
    end
  end
  
  def self.merge ( srcsent, destsent )
    case ((srcsent.is_a?(Array))? 2: 0) + ((destsent.is_a?(Array))? 1: 0)
    when 0 then # srcsent, destsent ɃubNIuWFNg
      table = { }
      srcsent.display do
        MergeMode::merge_src_block( srcsent, table )
        MergeMode::merge_dest_block( destsent, table )
      end
    when 1 then # destsent LbV
      table = { }
      srcsent.display do
        MergeMode::merge_src_block( srcsent, table )
        destsent.table.each do |name, image|
          $defout << image unless table.include?(name)
        end
      end
    when 2 then # srcsent LbV
      table = srcsent.table
      srcsent.sentblock.display do
        srcsent.each_value {|image| $defout << image }
        MergeMode::merge_dest_block( destsent, table )
      end
    when 3 then # srcsent, destsent ɃLbV
      if srcsent.is_a?(Array) and destsent.is_a?(Array) then
        destsent.table.update( srcsent.table )
        srcsent.sentblock.display do
          destsent.sentblock.each_value {|image| $defout << image }
        end
      end
    end
  end
end

def mode_merge ()
  v_print " + MODE = merge\n"
  
  # f[^̏
  srcfile = filename_to_stream($files[0])
  destfile = filename_to_stream($files[1])
  begin
    raise "Cannot open source file -- #{$file[0]}." unless srcfile
    raise "Cannot open destination file -- #{$file[1]}" unless destfile
    
    srccorpus = CorpusStream.new(srcfile).get_corpus()
    raise "No <pCorpus> in source file -- #{srcfile.path}." unless srccorpus
    v_print "   - Source = #{srcfile.path}\n"
    
    destcorpus = CorpusStream.new(destfile).get_corpus()
    raise "No <pCorpus> in destination file -- #{destfile.path}." unless destcorpus
    v_print "   - Destination = #{destfile.path}\n"
  rescue
    srcfile.close
    destfile.close
    fatal_exit $!
  end
  
  # 
  # ǂݍ񂾕...
  #  id  ... LbV瓯 id ̕
  #              ݂Ώo͂ăLbV폜
  # LbV̕...
  #  id v ... ǂ݂Ȃ merge 
  #  id sv ... LbVɊi[
  #  Е id Ȃ ... LbVɊi[
  #   id Ȃ ... ǂ݂Ȃ merge 
  srccorpus.display do
    srccache = { }
    destcache = { }
    srcsent = destsent = nil
    until srccorpus.closed? and destcorpus.closed? do
      srcsent = MergeMode::read_sentence(srccorpus) unless srcsent
      destsent = MergeMode::read_sentence(destcorpus) unless destsent
      next unless srcsent or destsent

      if !srcsent or !destsent then
        (srcsent || destsent).dump # Е̓ǂݍ݂ÎŎĉЕ͂̂܂܏o
        srcsent = destsent = nil
        next
      end

      cache = destcache.delete( String(srcsent['id']) )
      if cache then
        MergeMode::merge(srcsent, cache); srcsent = nil
        next
      end
      cache = srccache.delete( String(destsent['id']) )
      if cache then
        MergeMode::merge(destsent, cache); destsent = nil
        next
      end

      srcid = String(srcsent['id'])
      destid = String(destsent['id'])
      if (srcid.empty? and destid.empty?) or (srcid == destid) then
        #  id ݂Ȃ or  id v
        MergeMode::merge( srcsent, destsent )
        srcsent = destsent = nil
      else
        #  id vȂ or ǂ炩 id ݂Ȃ
        unless srcid.empty? then
          cache = srccache[srcid] = MergeMode::cache_sentence(srcsent)
          cache.id = srccache.size
          srcsent = nil
        end
        unless destid.empty? then
          cache = destcache[destid] = MergeMode::cache_sentence(destsent)
          cache.id = destcache.size
          destsent = nil
        end
      end
    end
    
    # LbV̒ǂo
    for cache in [srccache, destcache] do
      values = cache.values.sort {|a,b| a.id <=> b.id }
      for value in values do
        value.sentblock.display do
          value.table.each_value {|image| $defout << image }
        end
      end
    end
  end # srccorpus.display
end


#
#   Main Program
######################################################################

fetch_options(
  ['--count',  '-c', GetoptLong::REQUIRED_ARGUMENT],
  ['--width',  '-w', GetoptLong::REQUIRED_ARGUMENT],
  ['--prefix', '-p', GetoptLong::REQUIRED_ARGUMENT],
  ['--nocopy',       GetoptLong::NO_ARGUMENT],
  ['--sync',         GetoptLong::NO_ARGUMENT],
  ['--simply',       GetoptLong::NO_ARGUMENT]
)
$mode = $files.shift
case $mode
when "divide"
  ; # $defout g킯ł͂Ȃc
else
  $Verbose = false if $defout.tty?
end
v_print $VERSION

case $mode
when "extract" then
  mode_extract()
when "divide" then
  mode_divide()
when "select" then
  mode_select()
when "count" then
  mode_count()
when "sample" then
  mode_sample()
when "unite" then
  mode_unite()
when "analyze" then
  mode_analyze()
when "merge" then
  mode_merge()
else
  fatal_exit("Unknown mode ``#{$mode}''.") if String === $mode and $mode
  $stderr << $USAGE.call
end

