#!/usr/local/bin/ruby
# $Id: eval.rb,v 1.7 2002/11/01 05:59:43 abe Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   Precision evaluater
#
#      Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>
######################################################################
require File::join( File.dirname(__FILE__), "nlputil.rb" )


#
#   Used Exception Objects
######################################################################

class NoGrammarGivenException < RuntimeError
  # sɕKvȕ@^ĂȂꍇɔO
end


#
#   Core Object
######################################################################

class EvalUtility < NLPModule

  VERSION = <<"--------"
#{GEMINI::mklogo("Precision Evaluater   ver. 2.0")}
        Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>

--------

  USAGE = <<"--------"
usage:
  ruby eval.rb (options) (corpus file)

options:
  -g [grammar file]           using grammar

  --simply                    output result only

  --id STRING                 specify id attribute of <pCorpus>
--------

  # RXgN^
  #----------------------------------------
  def initialize ()
    super()
    @VERSION = VERSION
    @USAGE = USAGE
  end
  
  # IvVf[^CX^Xϐ
  #----------------------------------------
  def scan_argument ( argtbl )
    @GrammarFilename = String(argtbl['grammar'])
    
    @Simply = argtbl['simply'] || false
    
    @Id = String(argtbl['id'])
  end
  
  # [eBeBIuWFNg̎s
  #----------------------------------------
  def call_body ( argtbl )
    scan_argument( argtbl )
    @ResultLog = LogBlock.new()
    
    @total = @total_read = 0
    
    @grammar = read_grammar(@GrammarFilename)
    raise(NoGrammarGivenException, "no grammar given") if !@grammar
    
    attr_override = (@Id.empty?)? nil: {'id'=>@Id}
    process_multi_corpus(@inputs, attr_override) {|corpus| process_corpus(corpus) }
    fatal_exit("No sentences.") if @total == 0

    @ResultLog.table.each do |key, value|
      @ResultLog[key] =
      if Array === value then
        sprintf( "%.2f%%", 100.0 * value[0] / value[1] )
      else
        sprintf( "%.2f%%", 100.0 * value / @total )
      end
    end
    @ResultLog.display()
    v_print(" + OK.\n")

    v_print <<"END_OF_TEXT"
 + << Result >>
   - LT  = #{@ResultLog['LT']}
   - BT  = #{@ResultLog['BT']}
   - 0CB = #{@ResultLog['ZCB']}
   - LP  = #{@ResultLog['LP']}
   - LR  = #{@ResultLog['LR']}
   - BP  = #{@ResultLog['BP']}
   - BR  = #{@ResultLog['BR']}
   - CB  = #{@ResultLog['CB']}
END_OF_TEXT
    
    return self
  end
  
  def process_corpus ( corpus )
    print_progress = lambda { v_print( sprintf("   - \#%d (/%d)\r", @total, @total_read) ) }
    
    t = 0 ; print_progress.call()
    corpus.each do |sentence|
      t = update_progress_info( t ) { print_progress.call() }
      @total_read += 1
      begin
        sentence.display do
          answer, estimate, logblock = read_sentence( sentence )
          break unless answer and estimate # Ɨ\̗݂Ȃ΃_
          
          @total += 1
          localresult = eval_core( answer, estimate )
          
          localresult.each do |key, value|
            if Array === value then # LP Ȃ
              dat = Array( @ResultLog[key] )
              dat[0] = Integer(dat[0]) + value[0]
              dat[1] = Integer(dat[1]) + value[1]
            else # LT Ȃ
              dat = Integer( @ResultLog[key] )
              dat += value
            end
            @ResultLog[key] = dat
          end

          unless @Simply then
            localresult.each do |key, value|
              logblock[key] = (Array === value)? sprintf("%.2f%%", 100.0*value[0]/value[1]): value
            end
            logblock.display()
          end
        end
      rescue
        v_print("\n")
        print_warn(corpus.stream, "\##{@total}: #$!")
        t = 0
      end
    end #corpus.each
    print_progress.call() ; v_print("\n")
  end
  
  # <pSentence> ̓ǂݍŕKvȏW߂
  #----------------------------------------
  def read_sentence ( sentence )
    answer = nil
    estimate = nil
    logblock = LogBlock.new()

    out = (@Simply)? nil: $defout
    sentence.out = out
    sentence.each( TreeBlock, EstimateBlock, LogBlock ) do |cls, block|
      if cls then
        if cls == LogBlock then
          # <dLog> ͎Oŏo͂̂ō폜Ă
          block.out = nil
          logblock.update( str_to_hash(block.read) )
        elsif cls == TreeBlock then
          answer = block.display { TreeBlock.new(block).read(@grammar) }
        elsif cls == EstimateBlock then
          estimate = block.display { EstimateBlock.new(block).read(@grammar) }
        end
      else
        block.dump
      end
    end
    
    return [answer, estimate, logblock]
  end

  # ʏɊւe[u
  class BracketTable
    def initialize ()
      @table = [ ]
    end

    # left..right ̊ʂ݂邩𒲂ׂ
    def exist? ( range )
      $_ = @table[ range.first ]
      ($_)? $_[range.last]: false
    end

    # left [Ɏʂ̃Xg𓾂
    def at ( left )
      @table[left]
    end
    
    def store ( range )
      $_ = @table[ range.first ]
      $_ = @table[ range.first ] = [ ] unless $_
      $_[ range.last ] = true
    end

    # (left,right) ƌ銇ʂe[uɑ݂邩𒲂ׂ
    def cross_exist? ( range )
      left, right = range.first, range.last
      for i in (left+1)...right do
        next unless $_ = @table[i]
        for j in (right+1)...(@table.size) do
          next unless $_[j]
          return true
        end
      end
      return false
    end
  end

  def eval_core ( answer, estimate )
    localresult = { }

    # ̃xZbg
    ansary   = answer.glob()
    anscount = ansary.size # anscount: N_C

    # p̃ANZXe[u쐬
    anstbl = { }
    ansary.each {|elem| anstbl[elem] = 0 }

    ansbtbl = BracketTable.new()
    ansary.each {|elem| ansbtbl.store(elem[1]) }

    # {
    lt = bt = zcb = false
    lp = bp = 0.0 ; lpsrc = bpsrc = nil
    lr = br = 0.0 ; lrsrc = brsrc = nil
    cb = 0.0 ; cbsrc = nil
    estimate.each_with_index do |est, n|
      # \̃xZbg
      estary   = est.tree.glob()
      estcount = estary.size # estcount: N_G

      estbtbl = BracketTable.new()
      estary.each { |elem| estbtbl.store(elem[1]) }

      # \f[^𒆂Ɋ܂܂鐔𒲍
      common = 0
      estary.each { |elem| common += 1 if anstbl.include?(elem) }

      # ʈvvZp̃f[^𒊏o
      ansbcount = 0 # ansbcount: B_C
      ansary.each { |elem| ansbcount += 1 if estbtbl.exist?( elem[1] ) }
      estbcount = 0 # estbcount: B_G
      estary.each { |elem| estbcount += 1 if ansbtbl.exist?( elem[1] ) }

      # 0-CB vZp̃f[^𒊏o
      crosscount = 0
      estary.each { |elem| crosscount += 1 unless ansbtbl.cross_exist?( elem[1] ) }

      # \Ɛ̗Ɍ郉x̃xƈv LT
      if common == estcount and estcount == anscount then
        lt = true
      end

      # ʂ͈̔͂xv BT
      if ansbcount == anscount and anscount == estcount then
        bt = true
      end

      # Ȃʐ 0-CB
      if crosscount == estcount then
        zcb = true
      end

      # Precision  Recall ͍őlg
      $_ = 100.0 * common / estcount ; (lp = $_ ; lpsrc = [common, estcount]) if $_ > lp
      $_ = 100.0 * common / anscount ; (lr = $_ ; lrsrc = [common, anscount]) if $_ > lr

      $_ = 100.0 * estbcount / estcount ; (bp = $_ ; bpsrc = [estbcount, estcount]) if $_ > bp
      $_ = 100.0 * ansbcount / anscount ; (br = $_ ; brsrc = [ansbcount, anscount]) if $_ > br

      # CB őlg
      $_ = 100.0 * crosscount / estcount ; (cb = $_ ; cbsrc = [ crosscount, estcount ]) if $_ > cb
    end
    localresult.update(
      'LT'  => (lt )? 1: 0,
      'BT'  => (bt )? 1: 0,
      'ZCB' => (zcb)? 1: 0,
      'LP'  => lpsrc,
      'BP'  => bpsrc,
      'LR'  => lrsrc,
      'BR'  => brsrc,
      'CB'  => cbsrc
    )

    return localresult
  end

end

######################################################################
if $0 == __FILE__ then
### ̃t@Ĉ܂܎sꂽꍇ̏
print_error_message = lambda {|mes| $stderr << "ERROR: #{mes}\n" }

begin
  optspec = [ ]
  optspec << ['--grammar', '-g', ARG_RQ]
  optspec << ['--simply', ARG_NO]
  optspec << ['--id', ARG_RQ]
  
  core = EvalUtility.new()
  argtbl = { }
  inputs = core.fetch_options( ARGV, argtbl, optspec )
  if inputs then
    begin
      core.call( argtbl, inputs )
    rescue NoGrammarGivenException
      print_error_message.call( $! )
      core.print_usage( $stdout )
    end
  end
rescue
  print_error_message.call( $! )
end

###
end
######################################################################

