/*                                                     project Gemini 
----------------------------------------------------------------------
    Earley p[UIuWFNg
      $Id: eparser.c,v 1.12 2002/11/21 13:50:31 abe Exp $

       Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#define DEBUG_FREE  0  /**< IuWFNgɊւfobOx */
#define DEBUG_Tree  0  /**< Tree IuWFNgɊւfobOx */
#define DEBUG_OP  0  /**< ʑɊւfobOx */
#define DEBUG_PARSE  0  /**< \͂ɊւfobOx */
#define DEBUG_Prediction  0  /**< Prediction ɊւfobOx */
#define DEBUG_Scanning  0  /**< Scanning ɊւfobOx */
#define DEBUG_Completion  0  /**< Completion ɊւfobOx */
#include "nlputil.h"


/*  O[oϐȂ
------------------------------------------------------------------- */

static ID id_left;
static ID id_right;
static ID id_label;
static ID id_children;
static ID id_set_chart;
static ID id_get_chart;

static VALUE parse ( EParser* parser, Tree* tree, Chart* parentchart );


/*  ⏕p
------------------------------------------------------------------- */

static int subtree_left ( VALUE subtree )
{
    return NUM2INT( rb_funcall(subtree, id_left, 0) );
}

static int subtree_right ( VALUE subtree )
{
    return NUM2INT( rb_funcall(subtree, id_right, 0) );
}

static int subtree_children ( VALUE subtree )
{
    return rb_funcall( subtree, id_children, 0 );
}

static int subtree_children_length ( VALUE subtree )
{
    VALUE children = subtree_children( subtree );
    
    return RARRAY(children)->len; /* check_wordtree()  Array ł邱Ƃ͕ۏႳĂ */
}

static VALUE subtree_label ( VALUE subtree )
{
    return rb_funcall( subtree, id_label, 0 );
}


static VALUE check_wordtree ( VALUE wordtree )
{
    VALUE volatile children;
    VALUE volatile label;
    VALUE* ptr;
    int len;
    
    label = rb_funcall( wordtree, id_label, 0 );
    if( !NIL_P(label) ) {
        Check_Type( label, T_ARRAY );
        rb_ary_getptr( label, &ptr, &len );
        for( --len; len >= 0; --len ) {
            if( category_int(ptr[len]) == 0 ) {
                children = rb_String( rb_inspect(ptr[len]) );
                rb_raise( rb_eArgError, "wrong value of WordTree's label. (%s)", RSTRING(children)->ptr );
            }
        }
    }
    
    children = rb_funcall( wordtree, id_children, 0 );
    if( !NIL_P(children) ) {
        Check_Type( children, T_ARRAY );
        rb_ary_getptr( children, &ptr, &len );
        for( --len; len >= 0; --len )
            check_wordtree( ptr[len] );
    }
    
    return wordtree;
}

static void tree_new_core_debug ( int level )
{
    int j;
    
    fprintf( stderr, "[Tree] " );
    for( j = level; j > 0; --j )
        fprintf( stderr, " " );
}

static Tree* tree_new_core ( VALUE treetop )
{
    VALUE volatile children = rb_funcall( treetop, id_children, 0 );
    VALUE volatile label = rb_funcall( treetop, id_label, 0 );
    Tree* tree = ALLOC(Tree);
    VALUE* ptr;
    int len;
    int i;
    #if DEBUG_Tree > 0
    static int level = 0;
    #endif
    
    #if DEBUG_Tree > 0
    fprintf( stderr, "[Tree] addr=<%p>\n", tree );
    ++level;
    #endif
    
    /* f[^̐ݒ */
    tree->left = NUM2INT( rb_funcall( treetop, id_left, 0 ) );
    tree->right = NUM2INT( rb_funcall( treetop, id_right, 0 ) );
    #if DEBUG_Tree > 0
    tree_new_core_debug( level );
    fprintf( stderr, "left = %d, right = %d.\n", tree->left, tree->right );
    #endif
    
    #if DEBUG_Tree > 0
    tree_new_core_debug( level );
    #endif
    rb_ary_getptr( label, &ptr, &len );
    if( len > 0 ) {
        tree->label = ALLOC_N( int, len + 1 );
        for( i = 0; i < len; ++i ) {
            tree->label[i] = category_int( ptr[i] );
            #if DEBUG_Tree > 0
            fprintf( stderr, "label = %d.\n", tree->label[i] );
            #endif
        }
        tree->label[i] = 0;
    }
    else {
        tree->label = NULL;
        #if DEBUG_Tree > 0
        fprintf( stderr, "no label.\n", tree );
        #endif
    }
    
    #if DEBUG_Tree > 0
    tree_new_core_debug( level );
    #endif
    rb_ary_getptr( children, &ptr, &len );
    if( len > 0 ) {
        tree->child_len = len;
        tree->child_ary = ALLOC_N(Tree*, tree->child_len + 1);
        for( i = 0; i < tree->child_len; ++i )
            tree->child_ary[i] = tree_new( ptr[i] );
        tree->child_ary[i] = NULL;
        tree->prediction_label = (tree->child_ary[0]->label)? tree->child_ary[0]->label: tree->child_ary[0]->prediction_label;
    }
    else {
        tree->child_len = 0;
        tree->child_ary = NULL;
        tree->prediction_label = tree->label;
        #if DEBUG_Tree > 0
        fprintf( stderr, "no children.\n" );
        #endif
    }
    
    #if DEBUG_Tree > 0
    --level;
    #endif
    return tree;
}

Tree* tree_new ( VALUE tree )
{
    if( NIL_P(check_wordtree(tree)) )
        return NULL;
    return tree_new_core(tree);
}

void tree_free ( Tree* ptr )
{
    int i;
    
    if( ptr ) {
        for( i = 0; i < ptr->child_len; ++i )
            tree_free( ptr->child_ary[i] );
        ruby_xfree( ptr->label );
        ruby_xfree( ptr->child_ary );
        ruby_xfree( ptr );
    }
}


/*   \̓f[^
------------------------------------------------------------------- */

void parsedat_free ( ParseData* ptr )
{
    if( ptr ) {
        #if DEBUG_FREE > 0
        fprintf( stderr, "[FREE] parsedata_free(): begin.\n" );
        #endif
        ruby_xfree( ptr );
        #if DEBUG_FREE > 0
        fprintf( stderr, "[FREE] parsedata_free(): end.\n" );
        #endif
    }
}

void parsedat_mark ( ParseData* ptr )
{
    if( ptr ) {
        if( ptr->chart )
            rb_gc_mark( ptr->chart->self_object );
    }
}

VALUE parsedat_new ( EParser* parser, Tree* tree )
{
    VALUE volatile chartobj;
    VALUE volatile obj;
    ParseData* data;
    
    data = ALLOC_ZERO_N(ParseData, 1);
    obj = Data_Wrap_Struct( rb_cData, parsedat_mark, parsedat_free, data );
    data->self_object = obj;
    data->parser = parser;
    data->tree = tree;
    
    chartobj = chart_new( parser->start, parser->grammar, tree->left, tree->right ); /*$$ parser->start ͂ $$*/
    Data_Get_Struct( chartobj, Chart, data->chart );
    data->pred_table = ALLOC_ZERO_N( int*, tree->child_len );
    
    return obj;
}


/*   p[Ȕ
------------------------------------------------------------------- */

void eparser_free ( EParser* ptr )
{
    if( ptr ) {
        #if DEBUG_FREE > 0
        fprintf( stderr, "[FREE] eparser_free(): begin.\n" );
        #endif
        firstset_delete( ptr->firstset );
        ruby_xfree( ptr );
        #if DEBUG_FREE > 0
        fprintf( stderr, "[FREE] eparser_free(): end.\n" );
        #endif
    }
}

void eparser_mark ( EParser* ptr )
{
    if( ptr ) {
        rb_gc_mark( ptr->grammar->ruby_object );
        rb_gc_mark( ptr->start );
    }
}

static VALUE eparser_new ( VALUE klass, VALUE gramobj )
{
    VALUE volatile obj;
    EParser* parser;
    
    /* mۂ̓[NAĂ */
    parser = ALLOC( EParser );
    MEMZERO( parser, EParser, 1 );
    
    obj = Data_Wrap_Struct( klass, eparser_mark, eparser_free, parser );
    parser->self_object = obj;
    rb_obj_call_init( obj, 1, &gramobj );
    
    return obj;
}

static VALUE eparser_initialize ( VALUE self, VALUE gramobj )
{
    EParser* parser;
    Grammar* grammar;
    
    if( CLASS_OF(gramobj) != rb_cGrammar )
        rb_raise( rb_eArgError, "wrong type of argument. (requires Ext::Grammar)" );
    Data_Get_Struct( gramobj, Grammar, grammar );
    
    Data_Get_Struct( self, EParser, parser );
    parser->grammar = grammar;
    /* rb_obj_freeze( gramobj ) */
    parser->firstset = firstset_new( grammar );
    
    parser->start = Qnil;
    
    return self;
}


/*  \
------------------------------------------------------------------- */

/** @brief B\`FbN
 *
 * ^ꂽLAI[ƂȂLɓBł邩ǂ FIRST WgĔʂB
 *
 * @param fset [in] gp FIRST W
 * @param cat [in] o_ƂȂL
 * @param labels [in] I[ƂȂL (0 I[)
 * @retval 1 cat  labels ɓB\
 * @retval 0 Bs\
 */
static int is_reachable ( FIRSTSet const * fset, int cat, int const * labels )
{
    int* table;
    
    if( !fset )
        return 1;
    
    /* cat I[L̏ꍇ targets  cat Ɠꍇ̂ */
    if( cat <= 0 ) {
        for( ; *labels; ++labels ) {
            if( cat == *labels )
                return 1;
        }
        return 0;
    }
    
    /* B\f[^璲ׂ */
    table = (fset)? fset->ptr[cat]: NULL;
    if( !table )
        return 1; /* w肳ĂȂ̂œBłƍl */
    
    for( ; *labels; ++labels ) {
        if( table[*labels] )
            return 1;
    }
    return 0;
}

/** @brief Earley  Prediction Ń`[gɔzu
 *
 * w肳ꂽLӂɎ Earley  Prediction ŔzuB
 * zuʒu index ɂė^邪A̒ĺu݌Ă镔؂̂AԖڂ̎q؂ڎwvƂƂB
 * ̂߁Aindex Ɏw肷l̓`[g̈ʒuł͂Ȃ̂ŁAӂȂ΂ȂȂB
 *
 * ȂAzu鍀 Prediction Ȃ̂ŁARA͂̍łB
 * @ɓB\Xgݒ肳Ăꍇ́Aڎwq؂̏gĎ}sB
 *
 * @param data [in] \͂̋ʃf[^
 * @param cat [in] Ŏw肳ꂽLӂɎ Earley ǉ
 * @param index [in] ڕWƂ镔؂
 */
static void prediction_store ( ParseData* data, int cat, int index )
{
    FIRSTSet*  firstset;
    Grammar*  grammar = data->parser->grammar;
    Chart*  chart;
    Rule*  rule;
    Tree*  tree = data->tree;
    Tree*  child;
    State  tmpstate;
    int*  table;
    int  len, waiting;
    
    /* G[`FbN */
    if( cat == 0 )
        rb_raise( rb_eRangeError, "given null category for prediction" );
    
    if( !cfg_symbol_p( grammar, cat ) )
        rb_raise( rb_eRangeError, "given prediction symbol out of range (%d given for (%d,%d))", cat, -cfg_size_preterm(grammar), cfg_size_nonterm(grammar) );
    
    if( index < 0 || index >= tree->child_len )
        rb_raise( rb_eRangeError, "prediction index out of range (%d given for 0...%d)", index, tree->child_len );
    
    /* I[L͉oȂ */
    if( cat < 0 )
        return;
    
    /* ܂ĂȂ Prediction e[up */
    table = data->pred_table[index];
    if( !table ) {
        len = cfg_size_nonterm(grammar);
        table = data->pred_table[index] = ALLOC_ZERO_N(int, len);
    }
    
    /* ł Prediction sĂΉȂ (ǂ̂o邾) */
    if( table[cat] )
        return;
    table[cat] = 1;
    
    chart = data->chart;
    child = tree->child_ary[index];
    firstset = data->parser->firstset;
    if( is_reachable( firstset, cat, child->prediction_label ) ) {
        /* [ɓB\̂Ń`[g֓o^ */
        for( rule = grammar->hash_lhs[ cat ]; rule; rule = rule->next_lhs ) {
            /* ׂĂ̋Kǂ݂֓Bł킯ł͂Ȃ */
            waiting = rule->rhs[0];
            if( is_reachable( firstset, waiting, child->prediction_label ) ) {
                state_new( &tmpstate, rule, child->left );
                #if DEBUG_Prediction > 0
                fprintf( stderr, "[Prediction] %s\n", state_inspect(&tmpstate, grammar) );
                #endif
                chart_add_state2( chart, &tmpstate );
                
                /* ǉꂽ Prediction Ă */
                prediction_store( data, waiting, index );
            }
        }
    }
}

static State* _queue_shift ( ParseData* data )
{
    State* ret;
    
    if( !data->queue_head )
        return NULL;
    
    ret = data->queue_head;
    data->queue_head = ret->data;
    ret->data = NULL;
    
    if( !data->queue_head )
        data->queue_tail = NULL;
    
    return ret;
}

static void _queue_push ( ParseData* data, State* state )
{
    State* p;
    
    if( !data->queue_tail ) {
        data->queue_head = data->queue_tail = state;
        return;
    }
    
    for( p = data->queue_head; p; p = p->data ) {
        if( state_cmp( state, p ) == 0 )
            return;
    }
    data->queue_tail->data = state;
    data->queue_tail = state;
    state->data = NULL;
}

static void completion ( ParseData* data )
{
    Grammar* grammar = data->parser->grammar;
    Chart* chart = data->chart;
    State_p qstate, wstate, newstate;
    State statebuf;
    
    while( qstate = _queue_shift(data) ) {
        chart_add_state2( chart, qstate );
        for( wstate = chart_rwait( chart, qstate->top, qstate->left ); wstate; wstate = wstate->next_rwait ) {
            /* witem  qitem gݍ킹 */
            state_new3(&statebuf, wstate, qstate);
            if( !chart_find( chart, &statebuf ) ) {
                newstate = state_dup(NULL, &statebuf);
                #if DEBUG_Completion > 0
                fprintf( stderr, "[Completion]     %s\n", state_inspect(qstate, grammar) );
                fprintf( stderr, "[Completion]  +  %s\n", state_inspect(wstate, grammar) );
                fprintf( stderr, "[Completion] ==> %s\n", state_inspect(newstate, grammar) );
                #endif
                
                /* ̓L[ */
                if( state_complete_p( newstate ) )
                    _queue_push( data, newstate );
                else
                    chart_add_state2( chart, newstate );
            }
        }
    }
}

static void scanning_child ( ParseData* data, Tree* childtree )
{
    VALUE volatile childchart;
    EParser* parser = data->parser;
    Grammar* grammar = parser->grammar;
    Chart* parentchart = data->chart;
    Chart* chart;
    State_p state, nextstate;
    int i;
    
    childchart = parse( parser, childtree, parentchart );
    Data_Get_Struct( childchart, Chart, chart );
    
    for( i = chart->left; i <= chart->right; ++i ) {
        for( state = chart_dot(chart, i); state; state = nextstate ) {
            nextstate = state->next_dot;
            
            if( state->left == chart->left && state->right == chart->right && state_complete_p(state) ) {
                /* `[gŜɒ銮Ԃ̓L[։߂ */
                _queue_push( data, state );
            }
            else if( state->right == chart->right && !state_complete_p(state) ) {
                /* hbgʒu`[gE[ɂ関Ԃ͊Oƌ˂Ȃ̂Ŋ댯 */
                state_free( state );
            }
            else {
                /* e`[gֈړ */
                chart_add_state2( parentchart, state );
            }
        }
        *chart_dot_p(chart, i) = NULL;
    }
}

static void scanning ( ParseData* data, int index )
{
    Chart* chart = data->chart;
    Tree* child = data->tree->child_ary[index];
    State* newstate;
    int* catptr;
    
    /* ̍̂⊮ł̂𒲂ׂ */
    if( child->child_ary ) {
        /* q̍\͂s */
        scanning_child( data, child );
    }
    else {
        /* I[L scanning */
        for( catptr = child->label; *catptr; ++catptr ) {
            newstate = state_new2( NULL, child->left, child->right, *catptr );
            #if DEBUG_Scanning > 0
            fprintf( stderr, "[Scanning] %s\n", state_inspect(newstate, data->parser->grammar) );
            #endif
            _queue_push( data, newstate );
        }
    }
}

static void prediction ( ParseData* data, int index )
{
    Chart* chart = data->chart;
    Tree* child = data->tree->child_ary[index];
    State* state;
    
    /* hbgʒuɂL Prediction s */
    for( state = chart_dot( chart, child->left ); state; state = state->next_dot ) {
        if( state->rule && (state->dot < state->rule->rhslen) ) {
            /* p͖̂Ԃ */
            prediction_store( data, state->rule->rhs[state->dot], index );
        }
    }
}

static void set_goal ( ParseData* data, Chart* parent )
{
    Chart* chart = data->chart;
    State_p  state;
    Rule_p  rule;
    int*  labelptr;
    
    #if DEBUG_Prediction > 0
    fprintf( stderr, "[GOAL] Chart<%p>(%d,%d): set goal.\n", chart, chart->left, chart->right );
    #endif
    
    labelptr = data->tree->label;
    if( !labelptr ) {
        if( !parent ) {
            /* uxvuevȂ΃gbv */
            prediction_store( data, category_int(data->parser->start), 0 );
        }
        else {
            /* e`[gŃhbg[ʒuɂ鍀zu */
            for( state = chart_dot(parent, chart->left); state; state = state->next_dot ) {
                /* KłȂ΂ȂȂ */
                rule = state->rule;
                if( rule && (state->dot < rule->rhslen) )
                    prediction_store( data, rule->rhs[state->dot], 0 );
            }
        }
    }
    else {
        for( ; *labelptr; ++labelptr ) {
            /* `[g̍[ʒuexgbvɎzu */
            prediction_store( data, *labelptr, 0 );
        }
    }
}

static VALUE parse ( EParser* parser, Tree* tree, Chart* parentchart )
{
    VALUE volatile datobj;
    ParseData* dat;
    int i;
    
    datobj = parsedat_new( parser, tree );
    Data_Get_Struct( datobj, ParseData, dat );
    #if DEBUG_PARSE > 0
    fprintf( stderr, "[PARSE] parse(): (%d, %d)\n", dat->chart->left, dat->chart->right );
    #endif
    
    /* S[̐ݒ */
    set_goal( dat, parentchart );
    
    /* \ */
    for( i = 0; i < tree->child_len; ++i ) {
        prediction( dat, i );
        scanning( dat, i );
        completion( dat );
    }
    
    return dat->chart->self_object; /* KvȂ̂̓`[g */
}

VALUE eparser_parse ( VALUE self, VALUE wordtree )
{
    VALUE volatile chartobj = Qnil;
    VALUE volatile treeobj = Qnil;
    EParser* parser;
    Tree* tree;
    
    /* 󂯎̂ WordTree ̂ */
    if( CLASS_OF(wordtree) != rb_const_get( rb_cObject, rb_intern("WordTree") ) )
        rb_raise( rb_eArgError, "requires WordTree (given %s)", rb_class2name(CLASS_OF(wordtree)) );
    
    /* WordTree RpC */
    tree = tree_new( wordtree );
    treeobj = Data_Wrap_Struct( rb_cData, NULL, tree_free, tree );
    #if DEBUG_OP > 0
    fprintf( stderr, "[OP] eparser_parse(): WordTree conversion successed.\n" );
    #endif
    
    /* \͂ */
    Data_Get_Struct( self, EParser, parser );
    chartobj = parse( parser, tree, NULL );
    
    return chartobj;
}


/*   JnL̐ݒƎ擾
------------------------------------------------------------------- */

static VALUE eparser_set_start ( VALUE self, VALUE catobj )
{
    EParser* parser;
    int cat;
    
    cat = category_int( catobj );
    if( cat == 0 )
        rb_raise( rb_eArgError, "given invalid symbol as start symbol" );
    
    Data_Get_Struct( self, EParser, parser );
    
    parser->start = catobj;
    
    return catobj;
}

static VALUE eparser_get_start ( VALUE self )
{
    EParser* parser;
    
    Data_Get_Struct( self, EParser, parser );
    
    return parser->start;
}


/*   p[UNX̒`
------------------------------------------------------------------- */

void init_eparser ( void )
{
    VALUE  rb_cEParser;
    
    id_left = rb_intern("left");
    id_right = rb_intern("right");
    id_label = rb_intern("label");
    id_children = rb_intern("children");
    id_set_chart = rb_intern("chart=");
    id_get_chart = rb_intern("chart");
    
    /* Earley p[UNX̒` */
    rb_cEParser = rb_define_class_under( rb_mExt, "EarleyParser", rb_cData );
    
    /* NX\bh */
    rb_define_singleton_method( rb_cEParser, "new", eparser_new, 1 );
    
    /* \bh */
    rb_define_method( rb_cEParser, "initialize", eparser_initialize, 1 );
    rb_define_method( rb_cEParser, "parse", eparser_parse, 1 );
    rb_define_method( rb_cEParser, "start", eparser_get_start, 0 );
    rb_define_method( rb_cEParser, "start=", eparser_set_start, 1 );
}

