/*                                                     project Gemini 
----------------------------------------------------------------------
    mt@NX
      $Id: pcfg.c,v 1.4 2002/11/21 13:46:53 abe Exp $

       Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp> 
----------------------------------------------------------------------
*/
#define DEBUGSW 0
#include "nlputil.h"


/* Ruby ̃NXIuWFNg */
VALUE rb_cPCFG;
VALUE rb_cPCFG_Rule;


/*   [eBeB
------------------------------------------------------------------- */

static int pcfg_attr_to_flag ( PCFGAttr const * attr )
{
    int flag = 0;
    
    if( attr->lex != 0 )
        flag |= MODEL_LEXICALIZED;
    
    return flag;
}

static int pcfg_attr_cmp ( PCFGAttr const * attr1, PCFGAttr const * attr2 )
{
    if( attr1->flag != attr2->flag )
        return 0;
    
    if( attr1->lex != attr2->lex )
        return 0;
    
    return 1;
}


/*   mt@K̃fXgN^Ȃ
------------------------------------------------------------------- */

static void pcfgrule_free ( PCFGRule* r )
{
    if( r ) {
        ruby_xfree( r );
    }
}


/*   mt@K̕
------------------------------------------------------------------- */

static VALUE pcfgrule_to_s ( VALUE self )
{
    PCFG*     pcfg;
    PCFGRule* r;
    PCFGAttr* attr;
    Grammar*  gram;
    
    VALUE volatile cat;
    char buffer [4096];
    char *p;
    
    Data_Get_Struct( self, PCFGRule, r );
    p = buffer;
    
    p += sprintf( p, "%g ", r->prob );
    p = rule_typestring_internal( r->rule, p, -1 );
    
    attr = &r->attr;
    if( attr->flag != 0 ) {
        Data_Get_Struct( r->gramobj, PCFG, pcfg );
        Data_Get_Struct( pcfg->gramobj, Grammar, gram );
        p += sprintf( p, "|" );
        if( attr->flag & MODEL_LEXICALIZED ) {
            /* Lexicalized ̏ */
            cat = cfg_get_category( gram, attr->lex );
            p += sprintf( p, "lex=%s,", RSTRING(cat)->ptr );
        }
        *(--p) = '\0';
    }
    
    return rb_str_new2( buffer );
}

int pcfgrule_cmp ( PCFGRule const * r1, PCFGRule const * r2 )
{
    int  id1, id2;
    
    if( !r2 )
        return 1;
    
    id1 = r1->rule->id;
    id2 = r2->rule->id;
    if( id1 == id2 )
        return pcfg_attr_cmp( &(r1->attr), &(r2->attr) );
    
    return (id1 < id2)? -1: 1;
}

static VALUE pcfgrule_cmp_m ( VALUE self, VALUE obj )
{
    PCFGRule const * r1;
    PCFGRule const * r2;
    
    Data_Get_Struct( self, PCFGRule, r1 );
    if( CLASS_OF(obj) != rb_cPCFG_Rule )
        return rb_fix_new(1);
    Data_Get_Struct( self, PCFGRule, r2 );
    
    return rb_fix_new( pcfgrule_cmp(r1, r2) );
}


/*   mt@̐
----------------------------------------------------------------------
Ext::PCFG.new ( gram )

   @IuWFNg gram gĊmt@IuWFNg𐶐B
   ̊mt@IuWFNgɋKǉAgram ɂ̋K
Ȃ悤ȏꍇAgram ̓IɏĈŒӁB
*/

static void pcfg_free ( PCFG* pcfg )
{
    if( pcfg ) {
        ruby_xfree( pcfg );
    }
}

static void pcfg_mark ( PCFG* pcfg )
{
    if( pcfg ) {
        rb_gc_mark( pcfg->gramobj );
        rb_gc_mark( pcfg->ruleary );
    }
}

static VALUE pcfg_new ( VALUE klass, VALUE gramobj )
{
    VALUE volatile obj;
    PCFG* pcfg;
    
    pcfg = ALLOC_ZERO_N(PCFG, 1);
    obj = Data_Wrap_Struct( klass, pcfg_mark, pcfg_free, pcfg );
    rb_obj_call_init( obj, 1, &gramobj );
    
    return obj;
}

VALUE pcfg_create ( VALUE gramobj )
{
    return pcfg_new( rb_cPCFG, gramobj );
}

static VALUE pcfg_initialize ( VALUE self, VALUE gramobj )
{
    PCFG* pcfg;
    
    if( CLASS_OF(gramobj) != rb_cGrammar )
        rb_raise( rb_eNLPException, "Ext::ProbGram requires Ext::Grammar." );
    
    Data_Get_Struct( self, PCFG, pcfg );
    
    /* 킩Ă Ruby IuWFNgo^ */
    pcfg->ruby_object = self;
    pcfg->gramobj = gramobj;
    
    /* K̃Xg */
    pcfg->ruleary = rb_ary_new();
    
    return self;
}


/*   mt@ɋKǉ
----------------------------------------------------------------------
pcfg.store ( prob, rule )

   m prob 蓖ĂꂽK rule i[B
   rule ͎̏̒ʂB

(K) | (f) = (l), ...

   f͌݁A

 - lex

T|[gĂB
   rule \zɎw肳ꂽ gram ɑ݂ȂꍇAIɒǉB
*/

static char* pcfg_parse_model ( char* p, char* * modelstr, char* * valuestr )
{
    STRPARSE_SKIP_WS(p);
    if( *p == '\0' )
        return NULL;
    
    *modelstr = p;
    while( *p != ' ' && *p != '\t' && *p != '=' ) {
        if( *p == '\0' )
            rb_raise( rb_eNLPException, "PCFG rule: Syntax error. (no condition value)" );
        ++p;
    }
    STRPARSE_SKIP_WS(p);
    if( *p != '=' )
        rb_raise( rb_eNLPException, "PCFG rule: Syntax error." );
    ++p;
    STRPARSE_SKIP_WS(p);
    
    *valuestr = p;
    while( *p != ',' && *p != '\0' ) {
        if( *p == '\n' )
            *p = '\0';
        else
            ++p;
    }
    if( *p == ',' ) {
        *p = '\0';
        ++p;
    }
    
    return p; /* ̈ʒuԂ */
}

static void pcfg_store_parse ( PCFG* pcfg, PCFGRule* pcfgrule, char* buf )
{
    Category* cat;
    Grammar* gram;
    Rule* rule;
    
    VALUE volatile v;
    VALUE catobj;
    
    PCFGAttr attr;
    char *separator, *p, *model, *value;
    
    Data_Get_Struct( pcfg->gramobj, Grammar, gram );
    
    /* @K̉ */
    separator = strchr( buf, '|' );
    if( separator )
        *separator = '\0';
    cfg_add_rule_without_prop( gram, buf, &rule );
    if( rule->id < 0 )
        rb_raise( rb_eNLPException, "[BUG/nlputil.so] pcfg_store(): Rule-adding is failed." );
    pcfgrule->rule = rule;
    
    /* Lq̉ */
    MEMZERO( &attr, sizeof(PCFGAttr), 1 );
    if( separator ) {
        p = separator + 1; /* '|' ̒ォ͂n߂ */
        
        while( p = pcfg_parse_model( p, &model, &value ) ) {
            if( strcmp( model, "lex" ) == 0 ) { /* Lexicalized */
                if( attr.flag & MODEL_LEXICALIZED )
                    rb_raise( rb_eNLPException, "PCFG rule: Syntax error. (``lex'' is already declared)" );
                attr.flag |= MODEL_LEXICALIZED;
                v = rb_str_new2(value);
                
                catobj = v;
                cfg_add_symbol( gram, &catobj ); /* v ɂ͋LIuWFNg */
                Data_Get_Struct( catobj, Category, cat );
                attr.lex = cat->id;
            }
            else {
                rb_raise( rb_eNLPException, "PCFG rule: ``%s'' is unknown model.", model );
            }
        }
    }
    pcfgrule->attr = attr;
}

static VALUE pcfg_store ( VALUE self, VALUE probobj, VALUE strobj )
{
    PCFG* pcfg;
    VALUE volatile pcfgruleobj;
    PCFGRule pcfgrule, *pcfgruleptr;
    
    VALUE volatile v;
    char *buf;
    int ruleid;
    
    /* Kvȃf[^𑵂 */
    Data_Get_Struct( self, PCFG, pcfg );
    
    strobj = rb_String(strobj);
    buf = ALLOCA_N( char, RSTRING(strobj)->len + 1 ); /* X^bNɃRs[ strobj 𑀍삷 */
    MEMCPY( buf, RSTRING(strobj)->ptr, char, RSTRING(strobj)->len + 1 );
    
    pcfg_store_parse( pcfg, &pcfgrule, buf );
    pcfg_store_rule_ex( pcfg, pcfgrule.rule, &pcfgrule.attr, rb_num2dbl( rb_Float(probobj) ) );
    
    return self;
}

PCFG * pcfg_store_rule ( PCFG * pcfg, Rule * rule, double prob )
{
    return pcfg_store_rule_ex( pcfg, rule, NULL, prob );
}

PCFG * pcfg_store_rule_ex ( PCFG * pcfg, Rule * rule, PCFGAttr const * attr, double prob )
{
    PCFGRule * ruleptr;
    VALUE volatile ruleobj;
    VALUE volatile v;
    int ruleid;
    
    /* K̐ */
    ruleptr = ALLOC_ZERO_N( PCFGRule, 1 );
    ruleptr->gramobj = pcfg->ruby_object;
    ruleptr->rule = rule;
    ruleptr->prob = prob;
    if( attr )
        ruleptr->attr = *attr;
    
    ruleobj = Data_Wrap_Struct( rb_cPCFG_Rule, NULL, pcfgrule_free, ruleptr );
    ruleptr->ruby_object = ruleobj;
    
    /* @ւ̒ǉ */
    ruleid = rule->id;
    v = rb_ary_entry( pcfg->ruleary, ruleid );
    if( NIL_P(v) ) {
        v = rb_ary_new();
        rb_ary_store( pcfg->ruleary, ruleid, v );
    }
    if( TYPE(v) != T_ARRAY )
        rb_raise( rb_eTypeError, "[BUG/nlputil.so] pcfg_store(): pcfg->ruleary[%d] is not Array.", ruleid );
    
    if( RTEST(rb_ary_includes(v, ruleobj)) ) {
        v = pcfgrule_to_s( ruleobj );
        rb_raise( rb_eRuntimeError, "cannot add rule twice (%s)", RSTRING(v)->ptr );
    }
    rb_ary_push( v, ruleobj );
    
    return pcfg;
}


/*  PCFG IuWFNg̏o
----------------------------------------------------------------------
pcfg.display ( out )

   pcfg ̓e out ɏo͂B
   <dParameter> ubN`ŏo͂B
*/

static VALUE pcfg_display ( VALUE self, VALUE out )
{
    VALUE volatile rbstr_open_group = rb_str_new2("<dGroup>\n");
    VALUE volatile rbstr_close_group = rb_str_new2("</dGroup>\n");
    VALUE volatile rbstr_nl = rb_str_new2("\n");
    
    PCFG*     pcfg;
    PCFGRule* prule;
    Grammar*  gram;
    Rule_p    rule;
    
    VALUE ruleary;
    int   nonterm;
    int   lhs, i;
    
    /* -------------------- */
    
    Data_Get_Struct( self, PCFG, pcfg );
    Data_Get_Struct( pcfg->gramobj, Grammar, gram );
    nonterm = gram->size_nt;
    
    rb_io_write( out, rb_str_new2("<pParameter>\n") );
    for( lhs = 1; lhs < nonterm; ++lhs ) {
        rule = gram->hash_lhs[lhs];
        if( !rule )
            continue;
        
        rb_io_write( out, rbstr_open_group );
        for( ; rule; rule = rule->next_lhs ) {
            ruleary = rb_ary_entry( pcfg->ruleary, rule->id ); /*  rule->id ͈͊O */
            if( NIL_P(ruleary) )
                continue;
            
            for( i = 0; i < RARRAY(ruleary)->len; ++i ) {
                rb_io_write( out, RARRAY(ruleary)->ptr[i] );
                rb_io_write( out, rbstr_nl );
            }
        }
        rb_io_write( out, rbstr_close_group );
    }
    rb_io_write( out, rb_str_new2("</pParameter>\n") );
    
    return self;
}


/*   (t) ŋK̐m𓾂 (API)
------------------------------------------------------------------- */

static VALUE pcfg_get_ruleary ( PCFG const * pcfg, int ruleid, VALUE* *ptr, int* len )
{
    VALUE rary = rb_ary_entry( pcfg->ruleary, ruleid );
    
    if( NIL_P(rary) )
        rb_raise( rb_eRangeError, "[BUG/nlputil.so] pcfg_get_rule_probability(): ruleid out of range." );
    
    if( ptr )
        *ptr = RARRAY(rary)->ptr;
    
    if( len )
        *len = RARRAY(rary)->len;
    
    return rary;
}

double pcfg_get_rule_probability ( PCFG const * pcfg, int ruleid )
{
    PCFGRule* prule;
    VALUE volatile v;
    VALUE* ptr;
    int i, len;
    
    pcfg_get_ruleary(pcfg, ruleid, &ptr, &len);
    for( i = 0; i < len; ++i ) {
        Data_Get_Struct( ptr[i], PCFGRule, prule );
        if( prule->attr.flag != 0 )
            continue;
        return prule->prob;
    }
    
    v = rb_funcall( pcfg->gramobj, rb_intern("getrule"), 1, rb_fix_new(ruleid) );
    v = rb_funcall( v, rb_intern("typestring"), 0 );
    rb_raise( rb_eRuntimeError, "rule ``%s'' has no probability", RSTRING(v)->ptr );
    return 0.0; /* dummy */
}

double pcfg_get_rule_probability_ex ( PCFG const * pcfg, int ruleid, PCFGAttr* attr )
{
    PCFGRule* prule;
    VALUE volatile v;
    VALUE* ptr;
    int i, len;
    
    if( attr )
        attr->flag = pcfg_attr_to_flag( attr );
    
    if( !attr || attr->flag == 0 )
        return pcfg_get_rule_probability( pcfg, ruleid );
    
    pcfg_get_ruleary(pcfg, ruleid, &ptr, &len);
    for( i = 0; i < len; ++i ) {
        Data_Get_Struct( *ptr, PCFGRule, prule );
        ++ptr;
        if( !pcfg_attr_cmp( &prule->attr, attr ) )
            continue;
        return prule->prob;
    }
    
    v = rb_funcall( pcfg->gramobj, rb_intern("getrule"), 1, rb_fix_new(ruleid) );
    v = rb_funcall( v, rb_intern("typestring"), 0 );
    rb_raise( rb_eRuntimeError, "rule ``%s'' has no probability", RSTRING(v)->ptr );
    return 0.0; /* dummy */
}


VALUE pcfg_get_grammar ( VALUE self )
{
    PCFG*  pcfg;
    
    Data_Get_Struct( self, PCFG, pcfg );
    
    return pcfg->gramobj;
}


/*  ̊gW[̏R[h
------------------------------------------------------------------- */

void init_pcfg ( void )
{
    /* lCeBu PCFG NX̒`
    ---------------------------------------- */
    rb_cPCFG = rb_define_class_under( rb_mExt, "PCFG", rb_cData );
    rb_cPCFG_Rule = rb_define_class_under( rb_cPCFG, "Rule", rb_cData );
    
    /* NX\bh
    ---------------------------------------- */
    rb_define_singleton_method( rb_cPCFG, "new", pcfg_new, 1 );
    
    /* \bh (mt@)
    ---------------------------------------- */
    rb_define_method( rb_cPCFG, "initialize", pcfg_initialize, 1 );
    rb_define_method( rb_cPCFG, "store", pcfg_store, 2 );
    rb_define_method( rb_cPCFG, "display", pcfg_display, 1 );
    rb_define_method( rb_cPCFG, "grammar", pcfg_get_grammar, 0 );
    
    /* \bh (mtK)
    ---------------------------------------- */
    rb_define_method( rb_cPCFG_Rule, "to_s", pcfgrule_to_s, 0 );
    rb_define_method( rb_cPCFG_Rule, "<=>", pcfgrule_cmp_m, 1 );
}

