#!/usr/local/bin/ruby
# $Id: gem.rb,v 1.27 2002/11/22 02:59:47 abe Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   graphical EM
#
#      Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>
######################################################################
require File::join( File.dirname(__FILE__), "nlputil.rb" )
require File::join( File.dirname(__FILE__), "gem.so" ) unless defined? Ext::GraphicalEM


#
#   Used Exception Objects
######################################################################

class TooManyDataException < RuntimeError
  # f[^ꍇɔO
end

class NoDataGivenException < RuntimeError
  # wKɕKvȃf[^^ĂȂꍇɔO
end


#
#   Core Object
######################################################################

class GraphicalEMUtility < NLPModule
  
  DefaultThreshold = 0.000001
  DefaultLoopcount = 0
  DefaultDepth = 3
  
  # EM wK̃Of[^
  EMHistory = Struct.new(:loopcount, :likelihood, :time, :sec)
  
  VERSION = <<"--------"
#{GEMINI::mklogo("graphical EM algorithm   ver. 2.0")}
        Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>

--------

  USAGE = <<"--------"
usage:
  ruby gem.rb (options) (support graph files)

options:
  --param FILE                parameter file [required]
  -p FILE

  --count COUNT               iteration count
  -c COUNT

  --epsilon PRECISION         threshold to stop EM iteration
  -e PRECISION                  (default=#{DefaultThreshold})

  --depth SIZE                depth to search <dSupportGraph>
  -d SIZE                       (default=#{DefaultDepth})

  --random                    force to set random probability

--------
  
  # RXgN^
  #----------------------------------------
  def initialize ()
    super()
    @VERSION = VERSION
    @USAGE = USAGE
  end
  
  # IvVf[^CX^Xϐ
  #----------------------------------------
  def scan_argument ( argtbl )
    @ParamFilename = String(argtbl['param'])
    
    @Depth = argtbl['depth'].to_i
    @Depth = DefaultDepth if @Depth == 0
    
    arg = argtbl['epsilon']
    @Threshold = (arg)? Float(arg): DefaultThreshold
    @Threshold = nil if @Threshold == 0
    
    arg = argtbl['count']
    @Loopcount = (arg)? Integer(arg): DefaultLoopcount
    @Loopcount = nil if @Loopcount == 0
    
    @PrintInside = (argtbl['print-inside'])? true: false
    
    @PrintSg = (argtbl['print-sg'])? true: false
    
    @PrintParameter = (argtbl['print-param'])? true: false
    
    raise(ArgumentError, "specify condition to terminate EM loop") if !@Threshold and !@Loopcount
  end
  
  # [eBeBIuWFNg̎s
  #----------------------------------------
  def call_body ( argtbl )
    scan_argument( argtbl )
    
    param = read_param(@ParamFilename)
    raise(NoDataGivenException, "no parameter given") if !param
    
    # xOt̓o^
    gem = Ext::GraphicalEM.new( param )
    store_supportgraph( gem )
    
    # wKJn
    @start_time = Time.now()
    @start_ptime = get_process_time().ptime
    v_print(" + start EM learning\n")
    v_print("   - loop count = #{@Loopcount}\n") if @Loopcount
    v_print("   - threshold = #{@Threshold}\n") if @Threshold
    @history = [ ]
    while loop_continue?(gem) do
      gem.update()
      gem.display_sg(@stderr) if @PrintSg and defined?(gem.display_sg)
      gem.display_parameter(@stderr) if @PrintParameter and defined?(gem.display_parameter)
    end
    @end_time = Time.now()
    @end_ptime = get_process_time().ptime
    
    # wKʂ̏o
    gem.display(@stdout)
    
    # Ȍo
    display_log()
    
    return self
  end
  attr_reader :total  # ǂݍ񂾕̑
  
  def loop_continue? ( gem )
    ptime = get_process_time().ptime # [vJn̎ԂĂ
    
    # m̌vZs
    curr_likelihood = gem.likelihood()
    gem.display_inside(@stderr) if @PrintInside and defined?(gem.display_inside)
    
    @history << EMHistory.new(@history.length + 1, curr_likelihood, Time.now, ptime)
    v_print {
      hist = @history.last
      sec_delta = (@history.length > 1)? @history[-1].sec - @history[-2].sec: 0
      sprintf("   #%5d: (%s) %e <%f sec>\n", hist.loopcount, hist.time.strftime("%m/%d %H:%M"), hist.likelihood, sec_delta)
    }
    
    if @Threshold then
      if @history.length > 1 then
        prev_likelihood = @history[-2].likelihood
        return false if curr_likelihood - prev_likelihood < @Threshold
      end
    end
    
    if @Loopcount then
      return false if @history.length >= @Loopcount
    end
    
    return true
  end
  
  def display_log ()
    timeformat = "%Y/%m/%d %H:%M:%S"
    
    start_time = @start_time.strftime(timeformat)
    end_time = @end_time.strftime(timeformat)
    diff_time = @end_time.tv_sec - @start_time.tv_sec
    
    hostname = ENV['HOST'] || ENV['HOSTNAME'] || `hostname`.strip
    
    user = ENV['USER'] || ENV['USERNAME'] || `whoami`.strip
    
    interval = (@end_ptime - @start_ptime) / @history.length
    
    Block.display('dEMlog', nil, @stdout) do
      print "!user #{user}\n"
      print "!hostname #{hostname}\n"
      print "!time <#{start_time}> --- <#{end_time}> [#{diff_time} sec]\n"
      print "!interval #{interval}\n"
      print "!threshold #{@Threshold}\n" if @Threshold
      print "!loopcount #{@Loopcount}\n" if @Loopcount
      print "!history\n" \
            "\# loop\tprocess sec\tlikelihood\n"
      @history.each_with_index do |hist, i|
        proctime = (i > 0)? @history[i].sec - @history[i-1].sec: 0.0
        printf("%d\t%f\t%e\n", hist.loopcount, proctime, hist.likelihood)
      end
    end
  end
  
  # xOt̓o^
  #----------------------------------------
  def store_supportgraph ( gem )
    @total = 0
    @memtotal = 0
    
    sglist = [ ]
    @t = 0
    
    @inputs.each do |input|
      stream = filename_to_stream( input )
      if stream then
        find_supportgraph( sglist, Block.open(stream) )
        print_progress(); v_print EOL
      end
    end
    
    for sg in sglist do
      gem << sg
    end
  end
  
  def print_progress ()
    v_print {
      mem = Float(@memtotal) / 1024
      if mem > 1024 then
        sprintf( "    - \#%d (%.2f MB)\r", @total, mem/1024 )
      else
        sprintf( "    - \#%d (%.1f KB)\r", @total, mem )
      end
    }
  end
  
  def find_supportgraph ( sglist, parent )
    if SupportGraphBlock === parent then
      begin
        sglist << SupportGraphBlock.new(parent).read()
        @total += 1
        @memtotal += sglist.last.memsize
        @t = update_progress_info( @t ) { print_progress() }
      rescue
        v_print EOL; print_warn( parent, "\##{@total}: #{$!}" )
        @t = 0
      end
    else
      if PacketBlock === parent then
        if @Depth < 0 or parent.level+1 <= @Depth then
          parent.each do |block|
            find_supportgraph(sglist, block)
          end
        end
      end
    end
    parent.close()
  end
  
end


######################################################################
if $0 == __FILE__ then
### ̃t@Ĉ܂܎sꂽꍇ̏
print_error_message = lambda {|mes| $stderr << "ERROR: #{mes}\n" }

begin
  optspec = [ ]
  optspec << ['--param', '-p', ARG_RQ]
  optspec << ['--epsilon', '-e', ARG_RQ]
  optspec << ['--count', '-c', ARG_RQ]
  optspec << ['--depth', '-d', ARG_RQ]
  optspec << ['--random', ARG_RQ]
  optspec << ['--print-inside', ARG_NO] if $DEBUG
  optspec << ['--print-sg', ARG_NO] if $DEBUG
  optspec << ['--print-param', ARG_NO] if $DEBUG
  
  core = GraphicalEMUtility.new()
  argtbl = { }
  inputs = core.fetch_options( ARGV, argtbl, optspec )
  if inputs then
    begin
      core.call( argtbl, inputs )
    rescue NoDataGivenException
      print_error_message.call( $! )
      core.print_usage( $stdout )
    end
  end
rescue
  print_error_message.call( $! )
end

###
end
######################################################################

