#!/usr/local/bin/ruby
# $Id: viterbi.rb,v 1.7 2002/10/30 07:39:19 abe Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   Viterbi Parser
#
#      Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>
######################################################################
require File::join( File.dirname(__FILE__), "nlputil.rb" )


#
#   Global Method
######################################################################

# ftHgl
$nbest       =  1
$OPT_simply  =  false
$OPT_id      =  ""

$VERSION = <<"--------"
#{GEMINI::mklogo("Viterbi Parser   ver. 2.0")}
        Copyright (C) 2001-2002, SATO Lab. <http://mi.cs.titech.ac.jp>

--------

$USAGE = <<"--------"
usage:
  ruby viterbi.rb (options) (corpus file)

options:
  --grammar FILE              using grammar
  -g FILE

  --param FILE                parameter file
  -p FILE

  -n COUNT                    N-best parsing

  --simply                    output results only

  --id STRING                 specify id attribute of <pCorpus>
  
  --model MODEL               use PCFG extended model
--------


#
#   Main Program
######################################################################

fetch_options(
  ['--grammar', '-g', GetoptLong::REQUIRED_ARGUMENT],
  ['--param'  , '-p', GetoptLong::REQUIRED_ARGUMENT],
  ['-n'             , GetoptLong::REQUIRED_ARGUMENT],
  ['--simply'       , GetoptLong::NO_ARGUMENT],
  ['--id'           , GetoptLong::REQUIRED_ARGUMENT],
  ['--model'        , GetoptLong::REQUIRED_ARGUMENT]
)
v_print $VERSION
$Verbose = false if $defout.tty?

$nbest = Integer($OPT_n) if $OPT_n
fatal_exit("N-best out of range.") if $nbest <= 0 

# gp PCFG fɑΉ郁\bh𓾂
$method = :display_viterbi
if defined?($OPT_model) then
  $method = 
  case $OPT_model
  when "lex"
    method = :display_viterbi_lex
  else
    fatal_exit("Unknown PCFG model, ``#{$OPT_model}''.")
  end
end


#
#   Prepare grammar and PCFG
######################################################################

$grammar = read_grammar_file( $OPT_grammar )
$param = read_param_file( $OPT_param )

fatal_exit "No grammar." if !$grammar
fatal_exit "No parameter." if !$param or $param.empty?

$pcfg = Ext::PCFG.new( $grammar )
$param.each do |group|
  group.each do |pair|
    $pcfg.store( pair.first, pair.last )
  end
end

# $grammar  $pcfg т
$grammar.property['pcfg'] = $pcfg


#
#   Application Core
######################################################################

class ViterbiApplication < NLPApplication
  
  def initialize (*arg)
    super(*arg)
    @total = 0
    @totaltime = Struct::Tms.new( 0, 0, 0, 0 )
  end
  attr_reader :total      # ǂݍ񂾕̑
  attr_reader :totaltime  # vԂ̑v

  def print_progress ()
    v_print sprintf( "   - \#%d (%.1f sec)\r", @total, @totaltime.ptime )
  end
  
  def get_itemset ( block )
    itemset = nil
    if $OPT_simply then
      block.out = nil
      itemset = ItemSetBlock.new(block).read
    else
      # ł block ̏o͐ $defout ɂȂĂ
      block.display($defout) do
        itemset = ItemSetBlock.new(block).read
      end
    end
    block.close()
    return itemset
  end
  
  def read_sentence ( sentence )
    itemset = nil
    sentence.each( ItemSetBlock ) do |cls, block|
      if cls then
        raise NLPException, "Found too many <dItemSet>." if itemset
        itemset = get_itemset(block)
      else
        # <dEstimate> ͂̌Ő邽ߍ폜Ă
        # <dTree>  <dLog> ͌XgƎv̂ --simply ł폜Ȃ
        if EstimateBlock === block or ($OPT_simply and !(TreeBlock === block or LogBlock === block)) then
          block.skip
        else
          block.dump
        end
      end
    end
    return itemset
  end

  def on_execute ( corpus )
    t = 0 ; print_progress()
    corpus.each do |sentence|
      t = update_progress_info( t ) { print_progress() }
      @total += 1
      begin
        sentence.out = $defout
        sentence.display do
          # ̓ǂݍ
          itemset = read_sentence( sentence )
          break unless itemset
          
          # \
          display_block(EstimateBlock) do
            time_block { itemset.send( $method, $defout, $nbest ) }
            @totaltime += $time
          end
        end # sentence.display
      rescue
        print_warn( corpus.stream, "\##{@total}: #$!" )
        t = 0
      end
    end # corpus.each
    print_progress() ; v_print "\n"
  end
  
end


#
#   Parsing
######################################################################

app = ViterbiApplication.new( $files )

if $OPT_id.empty? then
  attr_override = nil
else
  attr_override = { 'id'=>$OPT_id }
end

v_print " + N-best = #{$nbest}\n"
app.execute( attr_override ) # \̖͂{

v_print " + OK.\n"
